#include "mail-message.h"
#include <iostream>
#include <sys/stat.h>
#include <strings.h>
#include <string.h>

MailMessage::MailMessage()
 : message_copied(false),
   have_received(false),
   have_bogosity(false),
   have_message_id(false),
   have_references(false),
   have_from_address(false),
   is_troll_message(false) {

   InitFromStdin();
}

MailMessage::~MailMessage() {
   DiscardStdin();
}

void
MailMessage::SetTrollFilter(bool init_is_troll_message) {
   is_troll_message = init_is_troll_message;
}

void
MailMessage::WriteToStdout() {

#ifdef BUILD_TROLLFILTER
   std::string troll;
   troll.append("X-TrollFilter: ");
   troll.append(is_troll_message ? "Yes" : "No");
   troll.append(endstr);
   header_lines.push_back(troll);
#endif

   int len = header_lines.size();
   std::string headerbuf;
   for (int i = 0; i < len; ++i) {
      headerbuf.append(header_lines[i]);
      headerbuf.append(endstr);
   }
   headerbuf.append(endstr);
   std::cout.write(headerbuf.c_str(), headerbuf.size());
   while (! std::cin.eof()) {
      char buffer[4096];
      std::cin.read(buffer, sizeof(buffer));
      std::streamsize amount = std::cin.gcount();
      if (amount > 0) {
         std::cout.write(buffer, amount);
      }
   }
   message_copied = true;
}

std::string
MailMessage::FromAddress() {
   if (! have_from_address) {
      std::string mid;
      if (SuckHeader("From:", mid, 0) != -1) {

         // Lets go for the one simplification that will cover 90%
         // of email addresses I see. Look for stuff between <>
         // characters.

         int lesschar = mid.find_first_of("<");
         if (lesschar >= 0) {
            int greatchar = mid.find_first_of(">", lesschar);
            if (greatchar >= lesschar) {
               std::string addrpart=
                              mid.substr(lesschar+1, greatchar-(lesschar+1));
               int spc = addrpart.find_first_of(" \t");
               int att = addrpart.find_first_of("@");
               if ((spc < 0) && (att > 0)) {

                  // Looks good. We have no space, we have an @, this
                  // is probably the email addr

                  mid = addrpart;
               }
            }
         }
         from_address = mid;

         // PUT CODE HERE - this could use simplification to just the
         // address using the more complex rules of mail header formats, but
         // as long as folks send mail the same way, the string will
         // probably still be the same from one message to another

      }
      have_from_address = true;
   }
   return from_address;
}

std::string
MailMessage::MessageID() {
   if (! have_message_id) {
      std::string mid;
      if (SuckHeader("Message-ID:", mid, 0) != -1) {
         message_id = mid;
      }
   }
   return message_id;
}

std::vector<std::string> &
MailMessage::References() {
   if (! have_references) {
      int len = header_lines.size();
      int i = 0;
      while ((i < len) && (i >= 0)) {
         std::string oneref;
         i = SuckHeader("References:", oneref, i);
         if ((i != -1) && (oneref.size() > 0)) {
            while (oneref.size() > 0) {
               int nonblank = oneref.find_first_not_of(" \t", 0);
               if (nonblank == std::string::npos) {
                  break;
               }
               int blank = oneref.find_first_of(" \t", nonblank);
               if (blank == std::string::npos) {
                  blank = oneref.size();
               }
               references.push_back(oneref.substr(nonblank, blank-nonblank));
               oneref.erase(0, blank);
            }
         }
      }
      have_references = true;
   }
   return references;
}

std::vector<std::string> &
MailMessage::Received() {
   if (! have_received) {
      int len = header_lines.size();
      int i = 0;
      while ((i < len) && (i >= 0)) {
         std::string onerec;
         i = SuckHeader("Received:", onerec, i);
         if ((i != -1) && (onerec.size() > 0)) {
            received.push_back(onerec);
         }
      }
      have_received = true;
   }
   return received;
}

std::string
MailMessage::Bogosity() {
   if (! have_bogosity) {
      int len = header_lines.size();
      for (int i = len - 1; (i >= 0) && (! have_bogosity); --i) {
         std::string line = header_lines[i];
         if ((line.size() >= 11) &&
             (strncasecmp(line.c_str(), "X-Bogosity:", 11) == 0)) {
            unsigned int nonspace = line.find_first_not_of(" \t", 11);
            bogosity.append(line.substr(nonspace));
            have_bogosity = true;
         }
      }
   }
   have_bogosity = true;
   return bogosity;
}

void
MailMessage::SetSpamIP(std::string ipaddr) {
   std::string newhdr("X-External-Spam: ");
   newhdr.append(ipaddr);
   header_lines.push_back(newhdr);
}

void
MailMessage::InitFromStdin() {

   endstr = "\n";

   // Read from stdin till we find a blank line and accumulate the
   // input in the "headers" string.

   bool use_crlf = false;
   bool first_line = true;
   while (std::cin.good()) {
      std::string str;
      getline(std::cin, str);
      if (str.size() > 0) {
         if (first_line) {

            // I'm not sure if the CR-LF line endings will be around
            // everywhere this filter is used, so check the first line
            // and remember which kind of line endings we saw so we can
            // be consistent when generating new lines on output.

            if (str[str.size()-1] == '\r') {
               use_crlf = true;
               endstr = "\r\n";
            } else {
               use_crlf = false;
            }
            first_line = false;
         }
         if (use_crlf && (str[str.size()-1] == '\r')) {
            str.erase(str.size()-1, 1);
         }
      }
      if (str.size() == 0) {

         // We have reached the end of the headers (a blank line).

         break;
      } else {

#ifdef BUILD_TROLLFILTER
         if (str.substr(0,14) == std::string("X-TrollFilter:")) {

            // Someone is trying to spoof an X-TrollFilter header.
            // Change this guy to X-SpoofedTrollFilter:

            str.insert(2,"Spoofed");
         }
#endif

         header_lines.push_back(str);
      }
   }
}

void
MailMessage::DiscardStdin() {
   if (message_copied) {
      return;
   }

   // I'm only doing this to prevent a broken pipe for my caller, but
   // if stdin is actually a regular file, I can avoid all the I/O...

   struct stat statbuf;
   if ((fstat(0, &statbuf) == 0) && S_ISREG(statbuf.st_mode)) {
      return;
   }
   while (! std::cin.eof()) {
      char buffer[4096];
      std::cin.read(buffer, sizeof(buffer));
   }
}

int
MailMessage::SuckHeader(
   const char * which_header,
   std::string & hdr,
   int start_line) {

   hdr.clear();
   int hdrlen = strlen(which_header);
   int len = header_lines.size();
   if ((start_line >= len) || (start_line < 0)) {
      return -1;
   }
   int found_hdr = -1;
   for (int i = start_line; i < len; ++i) {
      std::string line = header_lines[i];
      if ((line.size() >= hdrlen) &&
          (strncasecmp(line.c_str(), which_header, hdrlen) == 0)) {
         found_hdr = i;
         break;
      }
   }
   if (found_hdr == -1) {
      return found_hdr;
   }
   while (found_hdr < len) {
      std::string line = header_lines[found_hdr];
      int nonspace = line.find_first_not_of(" \t",hdrlen);
      if (nonspace == 0) {
         break;
      }
      if (nonspace > 0) {
         if (hdr.size() > 0) {
            hdr.append(" ");
         }
         hdr.append(line.substr(nonspace));
      }
      hdrlen = 0;
      ++found_hdr;
   }
   return found_hdr;
}
